"""
Allows setting a date range for when the page is active. Modifies the active()
manager method so that only pages inside the given range are used in the
default views and the template tags.

Depends on the page class having a "active_filters" list that will be used by
the page's manager to determine which entries are to be considered active.
"""
# ------------------------------------------------------------------------

from __future__ import absolute_import, unicode_literals

from datetime import datetime
from pytz.exceptions import AmbiguousTimeError

from django.db import models
from django.db.models import Q
from django.utils import timezone
from django.utils.cache import patch_response_headers
from django.utils.html import mark_safe
from django.utils.translation import ugettext_lazy as _

from feincms import extensions


# ------------------------------------------------------------------------
def format_date(d, if_none=''):
    """
    Format a date in a nice human readable way: Omit the year if it's the
    current year. Also return a default value if no date is passed in.
    """

    if d is None:
        return if_none

    now = timezone.now()
    fmt = (d.year == now.year) and '%d.%m' or '%d.%m.%Y'
    return d.strftime(fmt)


def latest_children(self):
    return self.get_children().order_by('-publication_date')


# ------------------------------------------------------------------------
def granular_now(n=None, default_tz=None):
    """
    A datetime.now look-alike that returns times rounded to a five minute
    boundary. This helps the backend database to optimize/reuse/cache its
    queries by not creating a brand new query each time.

    Also useful if you are using johnny-cache or a similar queryset cache.
    """
    if n is None:
        n = timezone.now()
    if default_tz is None:
        default_tz = n.tzinfo

    # Django 1.9:
    # The correct way to resolve the AmbiguousTimeError every dst
    # transition is... the is_dst parameter appeared with 1.9
    # make_aware(some_datetime, get_current_timezone(), is_dst=True)

    rounded_minute = (n.minute // 5) * 5
    d = datetime(n.year, n.month, n.day, n.hour, rounded_minute)
    try:
        retval = timezone.make_aware(d, default_tz)
    except AmbiguousTimeError:
        try:
            retval = timezone.make_aware(d, default_tz, is_dst=False)
        except TypeError:  # Pre-Django 1.9
            retval = timezone.make_aware(
                datetime(n.year, n.month, n.day, n.hour + 1, rounded_minute),
                default_tz)

    return retval


# ------------------------------------------------------------------------
def datepublisher_response_processor(page, request, response):
    """
    This response processor is automatically added when the datepublisher
    extension is registered. It sets the response headers to match with
    the publication end date of the page so that upstream caches and
    the django caching middleware know when to expunge the copy.
    """
    expires = page.publication_end_date
    if expires is not None:
        delta = expires - timezone.now()
        delta = int(delta.days * 86400 + delta.seconds)
        patch_response_headers(response, delta)


# ------------------------------------------------------------------------
class Extension(extensions.Extension):
    def handle_model(self):
        self.model.add_to_class(
            'publication_date',
            models.DateTimeField(_('publication date'), default=granular_now))
        self.model.add_to_class(
            'publication_end_date',
            models.DateTimeField(
                _('publication end date'),
                blank=True, null=True,
                help_text=_(
                    'Leave empty if the entry should stay active forever.')))
        self.model.add_to_class('latest_children', latest_children)

        # Patch in rounding the pub and pub_end dates on save
        orig_save = self.model.save

        def granular_save(obj, *args, **kwargs):
            if obj.publication_date:
                obj.publication_date = granular_now(obj.publication_date)
            if obj.publication_end_date:
                obj.publication_end_date = granular_now(
                    obj.publication_end_date)
            orig_save(obj, *args, **kwargs)
        self.model.save = granular_save

        # Append publication date active check
        if hasattr(self.model._default_manager, 'add_to_active_filters'):
            self.model._default_manager.add_to_active_filters(
                lambda queryset: queryset.filter(
                    Q(publication_date__lte=granular_now()) &
                     (Q(publication_end_date__isnull=True) |
                      Q(publication_end_date__gt=granular_now()))),
                key='datepublisher',
            )

        # Processor to patch up response headers for expiry date
        self.model.register_response_processor(
            datepublisher_response_processor)

    def handle_modeladmin(self, modeladmin):
        def datepublisher_admin(self, obj):
            return mark_safe('%s &ndash; %s' % (
                format_date(obj.publication_date),
                format_date(obj.publication_end_date, '&infin;'),
            ))
        datepublisher_admin.short_description = _('visible from - to')

        modeladmin.__class__.datepublisher_admin = datepublisher_admin

        try:
            pos = modeladmin.list_display.index('is_visible_admin')
        except ValueError:
            pos = len(modeladmin.list_display)

        modeladmin.list_display.insert(pos + 1, 'datepublisher_admin')

        modeladmin.add_extension_options(_('Date-based publishing'), {
            'fields': ['publication_date', 'publication_end_date'],
        })

# ------------------------------------------------------------------------



from __future__ import absolute_import, unicode_literals


# ------------------------------------------------------------------------
class ActiveAwareContentManagerMixin(object):
    """
    Implement what's necessary to add some kind of "active" state for content
    objects. The notion of active is defined by a number of filter rules that
    must all match (AND) for the object to be active.

    A Manager for a content class using the "datepublisher" extension
    should either adopt this mixin or implement a similar interface.
    """

    # A dict of filters which are used to determine whether a page is active or
    # not.  Extended for example in the datepublisher extension (date-based
    # publishing and un-publishing of pages). This will be set in
    # add_to_active_filters() below, so we won't share the same dict for
    # derived managers, do not replace with {} here!
    active_filters = None

    @classmethod
    def apply_active_filters(cls, queryset):
        """
        Apply all filters defined to the queryset passed and return the result.
        """
        if cls.active_filters is not None:
            for filt in cls.active_filters.values() or ():
                if callable(filt):
                    queryset = filt(queryset)
                else:
                    queryset = queryset.filter(filt)

        return queryset

    @classmethod
    def add_to_active_filters(cls, filter, key=None):
        """
        Add a new clause to the active filters. A filter may be either
        a Q object to be applied to the content class or a callable taking
        a queryset and spitting out a new one.

        If a filter with the given `key` already exists, the new filter
        replaces the old.
        """
        if cls.active_filters is None:
            cls.active_filters = {}
        if key is None:
            key = filter
        cls.active_filters[key] = filter

    def active(self):
        """
        Return only currently active objects.
        """
        return self.apply_active_filters(self)



"""
These are internal helpers. Do not rely on their presence.

http://mail.python.org/pipermail/python-dev/2008-January/076194.html
"""

from __future__ import absolute_import, unicode_literals

from distutils.version import LooseVersion
from django import get_version
from django.template.loader import render_to_string


__all__ = (
    'monkeypatch_method', 'monkeypatch_property',
)


def monkeypatch_method(cls):
    """
    A decorator to add a single method to an existing class::

        @monkeypatch_method(<someclass>)
        def <newmethod>(self, [...]):
            pass
    """

    def decorator(func):
        setattr(cls, func.__name__, func)
        return func
    return decorator


def monkeypatch_property(cls):
    """
    A decorator to add a single method as a property to an existing class::

        @monkeypatch_property(<someclass>)
        def <newmethod>(self, [...]):
            pass
    """

    def decorator(func):
        setattr(cls, func.__name__, property(func))
        return func
    return decorator


if LooseVersion(get_version()) < LooseVersion('1.10'):
    def ct_render_to_string(template, ctx, **kwargs):
        from django.template import RequestContext

        context_instance = kwargs.get('context')
        if context_instance is None and kwargs.get('request'):
            context_instance = RequestContext(kwargs['request'])

        return render_to_string(
            template,
            ctx,
            context_instance=context_instance)
else:
    def ct_render_to_string(template, ctx, **kwargs):
        return render_to_string(
            template,
            ctx,
            request=kwargs.get('request'))



"""
Add a keyword and a description field which are helpful for SEO optimization.
"""

from __future__ import absolute_import, unicode_literals

from django.db import models
from django.utils.translation import ugettext_lazy as _

from feincms import extensions


class Extension(extensions.Extension):
    def handle_model(self):
        self.model.add_to_class('meta_keywords', models.TextField(
            _('meta keywords'),
            blank=True,
            help_text=_('Keywords are ignored by most search engines.')))
        self.model.add_to_class('meta_description', models.TextField(
            _('meta description'),
            blank=True,
            help_text=_('This text is displayed on the search results page. '
                        'It is however not used for the SEO ranking. '
                        'Text longer than 140 characters is truncated.')))

    def handle_modeladmin(self, modeladmin):
        modeladmin.extend_list(
            'search_fields',
            ['meta_keywords', 'meta_description'],
        )

        modeladmin.add_extension_options(_('Search engine optimization'), {
            'fields': ('meta_keywords', 'meta_description'),
            'classes': ('collapse',),
        })



from __future__ import unicode_literals

from django.contrib import admin
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.translation import ugettext_lazy as _

from feincms._internal import ct_render_to_string
from feincms.admin.item_editor import FeinCMSInline
from feincms.module.medialibrary.fields import ContentWithMediaFile


class MediaFileContentInline(FeinCMSInline):
    raw_id_fields = ('mediafile',)
    radio_fields = {'type': admin.VERTICAL}


class MediaFileContent(ContentWithMediaFile):
    """
    Rehashed, backwards-incompatible media file content which does not contain
    the problems from v1 anymore.

    Create a media file content as follows::

        from feincms.content.medialibrary.models import MediaFileContent
        Page.create_content_type(MediaFileContent, TYPE_CHOICES=(
            ('default', _('Default')),
            ('lightbox', _('Lightbox')),
            ('whatever', _('Whatever')),
            ))

    For a media file of type 'image' and type 'lightbox', the following
    templates are tried in order:

    * content/mediafile/image_lightbox.html
    * content/mediafile/image.html
    * content/mediafile/lightbox.html
    * content/mediafile/default.html

    The context contains ``content`` and ``request`` (if available).
    """

    feincms_item_editor_inline = MediaFileContentInline

    class Meta:
        abstract = True
        verbose_name = _('media file')
        verbose_name_plural = _('media files')

    @classmethod
    def initialize_type(cls, TYPE_CHOICES=None):
        if TYPE_CHOICES is None:
            raise ImproperlyConfigured(
                'You have to set TYPE_CHOICES when'
                ' creating a %s' % cls.__name__)

        cls.add_to_class(
            'type',
            models.CharField(
                _('type'),
                max_length=20,
                choices=TYPE_CHOICES,
                default=TYPE_CHOICES[0][0],
            )
        )

    def render(self, **kwargs):
        return ct_render_to_string(
            [
                'content/mediafile/%s_%s.html' % (
                    self.mediafile.type, self.type),
                'content/mediafile/%s.html' % self.mediafile.type,
                'content/mediafile/%s.html' % self.type,
                'content/mediafile/default.html',
            ],
            {'content': self},
            request=kwargs.get('request'),
            context=kwargs.get('context'),
        )



from __future__ import absolute_import, unicode_literals

from collections import OrderedDict

from django import template
from django.contrib.auth import get_permission_codename


register = template.Library()


@register.simple_tag(takes_context=True)
def post_process_fieldsets(context, fieldset):
    """
    Removes a few fields from FeinCMS admin inlines, those being
    ``id``, ``DELETE`` and ``ORDER`` currently.

    Additionally, it ensures that dynamically added fields (i.e.
    ``ApplicationContent``'s ``admin_fields`` option) are shown.
    """
    # abort if fieldset is customized
    if fieldset.model_admin.fieldsets:
        return fieldset

    fields_to_include = set(fieldset.form.fields.keys())
    for f in ('id', 'DELETE', 'ORDER'):
        fields_to_include.discard(f)

    def _filter_recursive(fields):
        ret = []
        for f in fields:
            if isinstance(f, (list, tuple)):
                # Several fields on one line
                sub = _filter_recursive(f)
                # Only add if there's at least one field left
                if sub:
                    ret.append(sub)
            elif f in fields_to_include:
                ret.append(f)
                fields_to_include.discard(f)
        return ret

    new_fields = _filter_recursive(fieldset.fields)
    # Add all other fields (ApplicationContent's admin_fields) to
    # the end of the fieldset
    for f in fields_to_include:
        new_fields.append(f)

    if context.get('request'):
        new_fields.extend(list(
            fieldset.model_admin.get_readonly_fields(
                context.get('request'),
                context.get('original'),
            )
        ))

    fieldset.fields = new_fields
    return ''


@register.inclusion_tag('admin/feincms/content_type_selection_widget.html',
                        takes_context=True)
def show_content_type_selection_widget(context, region):
    """
    {% show_content_type_selection_widget region %}
    """
    user = context['request'].user
    types = OrderedDict({None: []})

    for ct in region._content_types:
        # Skip cts that we shouldn't be adding anyway
        opts = ct._meta
        perm = opts.app_label + "." + get_permission_codename('add', opts)
        if not user.has_perm(perm):
            continue

        types.setdefault(
            getattr(ct, 'optgroup', None),
            [],
        ).append((ct.__name__.lower, ct._meta.verbose_name))

    return {'types': types}



# ------------------------------------------------------------------------
# coding=utf-8
# ------------------------------------------------------------------------

from __future__ import absolute_import, unicode_literals

import logging

import django
from django import template
from django.conf import settings
from django.utils.safestring import mark_safe

from feincms.utils import get_singleton, get_singleton_url


register = template.Library()
assignment_tag = (
    register.simple_tag if django.VERSION >= (1, 9)
    else register.assignment_tag
)


def _render_content(content, **kwargs):
    # Track current render level and abort if we nest too deep. Avoids
    # crashing in recursive page contents (eg. a page list that contains
    # itself or similar).
    request = kwargs.get('request')
    if request is not None:
        level = getattr(request, 'feincms_render_level', 0)
        if level > 10:
            logging.getLogger('feincms').error(
                'Refusing to render %r, render level is already %s' % (
                    content, level))
            return
        setattr(request, 'feincms_render_level', level + 1)

    r = content.render(**kwargs)

    if request is not None:
        level = getattr(request, 'feincms_render_level', 1)
        setattr(request, 'feincms_render_level', max(level - 1, 0))

    return r


@register.simple_tag(takes_context=True)
def feincms_render_region(context, feincms_object, region, request=None):
    """
    {% feincms_render_region feincms_page "main" request %}
    """
    if not feincms_object:
        return ''

    return mark_safe(''.join(
        _render_content(content, request=request, context=context)
        for content in getattr(feincms_object.content, region)))


@register.simple_tag(takes_context=True)
def feincms_render_content(context, content, request=None):
    """
    {% feincms_render_content content request %}
    """
    if not content:
        return ''

    return _render_content(content, request=request, context=context)


@assignment_tag
def feincms_load_singleton(template_key, cls=None):
    """
    {% feincms_load_singleton template_key %} -- return a FeinCMS
    Base object which uses a Template with singleton=True.
    """
    return get_singleton(template_key, cls, raise_exception=settings.DEBUG)


@register.simple_tag
def feincms_singleton_url(template_key, cls=None):
    """
    {% feincms_singleton_url template_key %} -- return the URL of a FeinCMS
    Base object which uses a Template with singleton=True.
    """
    return get_singleton_url(template_key, cls, raise_exception=settings.DEBUG)



from __future__ import absolute_import, unicode_literals

from django.db import models
from django.utils.translation import ugettext_lazy as _

from feincms._internal import ct_render_to_string
from feincms.content.raw.models import RawContent  # noqa
from feincms.content.richtext.models import RichTextContent  # noqa


class TemplateContent(models.Model):
    """
    Pass a list of templates when creating this content type. It uses the
    default template system::

        Page.create_content_type(TemplateContent, TEMPLATES=[
            ('content/template/something1.html', 'something'),
            ('content/template/something2.html', 'something else'),
            ('base.html', 'makes no sense'),
        ])
    """
    class Meta:
        abstract = True
        verbose_name = _('template content')
        verbose_name_plural = _('template contents')

    @classmethod
    def initialize_type(cls, TEMPLATES):
        cls.add_to_class('template', models.CharField(
            _('template'),
            max_length=100,
            choices=TEMPLATES,
        ))

    def render(self, **kwargs):
        return ct_render_to_string(
            self.template,
            {'content': self},
            request=kwargs.get('request'),
            context=kwargs.get('context'),
        )



"""
Extend or modify the navigation with custom entries.

This extension allows the website administrator to select an extension
which processes, modifies or adds subnavigation entries. The bundled
``feincms_nav`` template tag knows how to collect navigation entries,
be they real Page instances or extended navigation entries.
"""

from __future__ import absolute_import, unicode_literals

from collections import OrderedDict
import types

from django.db import models
from django.utils import six
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from feincms import extensions
from feincms.utils import get_object, shorten_string
from feincms._internal import monkeypatch_method


class TypeRegistryMetaClass(type):
    """
    You can access the list of subclasses as <BaseClass>.types

    TODO use NavigationExtension.__subclasses__() instead?
    """

    def __init__(cls, name, bases, attrs):
        if not hasattr(cls, 'types'):
            cls.types = []
        else:
            cls.types.append(cls)


class PagePretender(object):
    """
    A PagePretender pretends to be a page, but in reality is just a shim layer
    that implements enough functionality to inject fake pages eg. into the
    navigation tree.

    For use as fake navigation page, you should at least define the following
    parameters on creation: title, url, level. If using the translation
    extension, also add language.
    """
    pk = None

    # emulate mptt properties to get the template tags working
    class _mptt_meta:
        level_attr = 'level'

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def get_absolute_url(self):
        return self.url

    def get_navigation_url(self):
        return self.get_absolute_url()

    def get_level(self):
        return self.level

    def get_children(self):
        """ overwrite this if you want nested extensions using recursetree """
        return []

    def available_translations(self):
        return ()

    def get_original_translation(self, page):
        return page

    def short_title(self):
        return shorten_string(self.title)


class NavigationExtension(six.with_metaclass(TypeRegistryMetaClass)):
    """
    Base class for all navigation extensions.

    The name attribute is shown to the website administrator.
    """

    name = _('navigation extension')

    def children(self, page, **kwargs):
        """
        This is the method which must be overridden in every navigation
        extension.

        It receives the page the extension is attached to, the depth up to
        which the navigation should be resolved, and the current request object
        if it is available.
        """

        raise NotImplementedError


def navigation_extension_choices():
    for ext in NavigationExtension.types:
        if (issubclass(ext, NavigationExtension) and
                ext is not NavigationExtension):
            yield ('%s.%s' % (ext.__module__, ext.__name__), ext.name)


def get_extension_class(extension):
    extension = get_object(extension)
    if isinstance(extension, types.ModuleType):
        return getattr(extension, 'Extension')
    return extension


class Extension(extensions.Extension):
    ident = 'navigation'  # TODO actually use this
    navigation_extensions = None

    @cached_property
    def _extensions(self):
        if self.navigation_extensions is None:
            return OrderedDict(
                ('%s.%s' % (ext.__module__, ext.__name__), ext)
                for ext in NavigationExtension.types
                if (
                    issubclass(ext, NavigationExtension) and
                    ext is not NavigationExtension))

        else:
            return OrderedDict(
                ('%s.%s' % (ext.__module__, ext.__name__), ext)
                for ext
                in map(get_extension_class, self.navigation_extensions))

    def handle_model(self):
        choices = [
            (path, ext.name) for path, ext in self._extensions.items()]

        self.model.add_to_class(
            'navigation_extension',
            models.CharField(
                _('navigation extension'),
                choices=choices,
                blank=True, null=True, max_length=200,
                help_text=_(
                    'Select the module providing subpages for this page if'
                    ' you need to customize the navigation.')))

        extension = self

        @monkeypatch_method(self.model)
        def extended_navigation(self, **kwargs):
            if not self.navigation_extension:
                return self.children.in_navigation()

            cls = None

            try:
                cls = extension._extensions[self.navigation_extension]
            except KeyError:
                cls = get_object(self.navigation_extension, fail_silently=True)
                extension._extensions[self.navigation_extension] = cls

            if cls:
                return cls().children(self, **kwargs)
            return self.children.in_navigation()

    def handle_modeladmin(self, modeladmin):
        modeladmin.add_extension_options(_('Navigation extension'), {
            'fields': ('navigation_extension',),
            'classes': ('collapse',),
        })



# flake8: noqa
from __future__ import absolute_import, unicode_literals

import warnings

from feincms.extensions.datepublisher import *

warnings.warn(
    'Import %(name)s from feincms.extensions.%(name)s' % {
        'name': __name__.split('.')[-1],
    }, DeprecationWarning, stacklevel=2)



from __future__ import absolute_import, unicode_literals

from django.db import models
from django.utils.translation import ugettext_lazy as _

from feincms import settings
from feincms._internal import ct_render_to_string
from feincms.contrib.richtext import RichTextField


class RichTextContent(models.Model):
    """
    Rich text content. Uses TinyMCE by default, but can be configured to do
    anything you want using ``FEINCMS_RICHTEXT_INIT_CONTEXT`` and
    ``FEINCMS_RICHTEXT_INIT_TEMPLATE``.

    If you are using TinyMCE 4.x then ``FEINCMS_RICHTEXT_INIT_TEMPLATE``
    needs to be set to ``admin/content/richtext/init_tinymce4.html``.

    Optionally runs the HTML code through HTML cleaners if you specify
    ``cleanse=True`` when calling ``create_content_type``.
    """

    feincms_item_editor_context_processors = (
        lambda x: settings.FEINCMS_RICHTEXT_INIT_CONTEXT,
    )
    feincms_item_editor_includes = {
        'head': [settings.FEINCMS_RICHTEXT_INIT_TEMPLATE],
    }

    class Meta:
        abstract = True
        verbose_name = _('rich text')
        verbose_name_plural = _('rich texts')

    def render(self, **kwargs):
        return ct_render_to_string(
            'content/richtext/default.html',
            {'content': self},
            request=kwargs.get('request'),
            context=kwargs.get('context'),
        )

    @classmethod
    def initialize_type(cls, cleanse=None):
        cls.add_to_class(
            'text',
            RichTextField(_('text'), blank=True, cleanse=cleanse),
        )



from __future__ import absolute_import, print_function, unicode_literals

import logging
import re
import sys

from django.conf import settings as django_settings
from django.http import Http404, HttpResponseRedirect


logger = logging.getLogger(__name__)


def redirect_request_processor(page, request):
    """
    Returns a ``HttpResponseRedirect`` instance if the current page says
    a redirect should happen.
    """
    target = page.get_redirect_to_target(request)
    if target:
        extra_path = request._feincms_extra_context.get('extra_path', '/')
        if extra_path == '/':
            return HttpResponseRedirect(target)
        logger.debug(
            "Page redirect on '%s' not taken because extra path '%s' present",
            page.get_absolute_url(), extra_path)
        raise Http404()


def extra_context_request_processor(page, request):
    """
    Fills ``request._feincms_extra_context`` with a few useful variables.
    """
    request._feincms_extra_context.update({
        # XXX This variable name isn't accurate anymore.
        'in_appcontent_subpage': False,
        'extra_path': '/',
    })

    url = page.get_absolute_url()
    if request.path != url:
        request._feincms_extra_context.update({
            'in_appcontent_subpage': True,
            'extra_path': re.sub(
                '^' + re.escape(url.rstrip('/')),
                '',
                request.path,
            ),
        })


def etag_request_processor(page, request):
    """
    Short-circuits the request-response cycle if the ETag matches.
    """

    # XXX is this a performance concern? Does it create a new class
    # every time the processor is called or is this optimized to a static
    # class??
    class DummyResponse(dict):
        """
        This is a dummy class with enough behaviour of HttpResponse so we
        can use the condition decorator without too much pain.
        """
        def has_header(page, what):
            return False

    def dummy_response_handler(*args, **kwargs):
        return DummyResponse()

    def etagger(request, page, *args, **kwargs):
        etag = page.etag(request)
        return etag

    def lastmodifier(request, page, *args, **kwargs):
        lm = page.last_modified()
        return lm

    # Unavailable in Django 1.0 -- the current implementation of ETag support
    # requires Django 1.1 unfortunately.
    from django.views.decorators.http import condition

    # Now wrap the condition decorator around our dummy handler:
    # the net effect is that we will be getting a DummyResponse from
    # the handler if processing is to continue and a non-DummyResponse
    # (should be a "304 not modified") if the etag matches.
    rsp = condition(etag_func=etagger, last_modified_func=lastmodifier)(
        dummy_response_handler)(request, page)

    # If dummy then don't do anything, if a real response, return and
    # thus shortcut the request processing.
    if not isinstance(rsp, DummyResponse):
        return rsp


def etag_response_processor(page, request, response):
    """
    Response processor to set an etag header on outgoing responses.
    The Page.etag() method must return something valid as etag content
    whenever you want an etag header generated.
    """
    etag = page.etag(request)
    if etag is not None:
        response['ETag'] = '"' + etag + '"'


def debug_sql_queries_response_processor(verbose=False, file=sys.stderr):
    """
    Attaches a handler which prints the query count (and optionally all
    individual queries which have been executed) on the console. Does nothing
    if ``DEBUG = False``.

    Example::

        from feincms.module.page import models, processors
        models.Page.register_response_processor(
            processors.debug_sql_queries_response_processor(verbose=True),
            )
    """
    if not django_settings.DEBUG:
        return lambda page, request, response: None

    def processor(page, request, response):
        from django.db import connection

        try:
            import sqlparse

            def print_sql(x):
                return sqlparse.format(
                    x, reindent=True, keyword_case='upper')
        except Exception:
            def print_sql(x):
                return x

        if verbose:
            print("-" * 60, file=file)
        time = 0.0
        i = 0
        for q in connection.queries:
            i += 1
            if verbose:
                print("%d : [%s]\n%s\n" % (
                    i, q['time'], print_sql(q['sql'])), file=file)
            time += float(q['time'])

        print("-" * 60, file=file)
        print("Total: %d queries, %.3f ms" % (i, time), file=file)
        print("-" * 60, file=file)

    return processor



"""
Page navigation groups allow assigning pages to differing navigation lists
such as header, footer and what else.
"""

from __future__ import absolute_import, unicode_literals

from django.db import models
from django.utils.translation import ugettext_lazy as _

from feincms import extensions


class Extension(extensions.Extension):
    ident = 'navigationgroups'
    groups = [
        ('default', _('Default')),
        ('footer', _('Footer')),
    ]

    def handle_model(self):
        self.model.add_to_class(
            'navigation_group',
            models.CharField(
                _('navigation group'),
                choices=self.groups,
                default=self.groups[0][0],
                max_length=20,
                blank=True,
                db_index=True))

    def handle_modeladmin(self, modeladmin):
        modeladmin.add_extension_options('navigation_group')
        modeladmin.extend_list('list_display', ['navigation_group'])
        modeladmin.extend_list('list_filter', ['navigation_group'])



"""
Add a many-to-many relationship field to relate this page to other pages.
"""

from __future__ import absolute_import, unicode_literals

from django.db import models
from django.utils.translation import ugettext_lazy as _

from feincms import extensions, settings


class Extension(extensions.Extension):
    def handle_model(self):
        self.model.add_to_class('related_pages', models.ManyToManyField(
            settings.FEINCMS_DEFAULT_PAGE_MODEL,
            blank=True,
            related_name='%(app_label)s_%(class)s_related',
            help_text=_(
                'Select pages that should be listed as related content.')))

    def handle_modeladmin(self, modeladmin):
        modeladmin.extend_list('filter_horizontal', ['related_pages'])

        modeladmin.add_extension_options(_('Related pages'), {
            'fields': ('related_pages',),
            'classes': ('collapse',),
        })



# ------------------------------------------------------------------------
# coding=utf-8
# ------------------------------------------------------------------------

from __future__ import absolute_import, unicode_literals

import os
import re

from django.db import models
from django.db.models.signals import post_delete
from django.dispatch.dispatcher import receiver
from django.template.defaultfilters import slugify
from django.utils import timezone
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from feincms import settings
from feincms.models import ExtensionsMixin
from feincms.translations import (
    TranslatedObjectMixin, Translation, TranslatedObjectManager)

from . import logger


# ------------------------------------------------------------------------
class CategoryManager(models.Manager):
    """
    Simple manager which exists only to supply ``.select_related("parent")``
    on querysets since we can't even __str__ efficiently without it.
    """
    def get_queryset(self):
        return super(CategoryManager, self).get_queryset().select_related(
            "parent")


# ------------------------------------------------------------------------
@python_2_unicode_compatible
class Category(models.Model):
    """
    These categories are meant primarily for organizing media files in the
    library.
    """

    title = models.CharField(_('title'), max_length=200)
    parent = models.ForeignKey(
        'self', blank=True, null=True,
        on_delete=models.CASCADE,
        related_name='children', limit_choices_to={'parent__isnull': True},
        verbose_name=_('parent'))

    slug = models.SlugField(_('slug'), max_length=150)

    class Meta:
        ordering = ['parent__title', 'title']
        verbose_name = _('category')
        verbose_name_plural = _('categories')
        app_label = 'medialibrary'

    objects = CategoryManager()

    def __str__(self):
        if self.parent_id:
            return '%s - %s' % (self.parent.title, self.title)

        return self.title

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = slugify(self.title)

        super(Category, self).save(*args, **kwargs)
    save.alters_data = True

    def path_list(self):
        if self.parent is None:
            return [self]
        p = self.parent.path_list()
        p.append(self)
        return p

    def path(self):
        return ' - '.join((f.title for f in self.path_list()))


# ------------------------------------------------------------------------
@python_2_unicode_compatible
class MediaFileBase(models.Model, ExtensionsMixin, TranslatedObjectMixin):
    """
    Abstract media file class. Includes the
    :class:`feincms.models.ExtensionsMixin` because of the (handy) extension
    mechanism.
    """

    file = models.FileField(
        _('file'), max_length=255,
        upload_to=settings.FEINCMS_MEDIALIBRARY_UPLOAD_TO)
    type = models.CharField(
        _('file type'), max_length=12, editable=False,
        choices=())
    created = models.DateTimeField(
        _('created'), editable=False, default=timezone.now)
    copyright = models.CharField(_('copyright'), max_length=200, blank=True)
    file_size = models.IntegerField(
        _("file size"), blank=True, null=True, editable=False)

    categories = models.ManyToManyField(
        Category, verbose_name=_('categories'), blank=True)
    categories.category_filter = True

    class Meta:
        abstract = True
        ordering = ['-created']
        verbose_name = _('media file')
        verbose_name_plural = _('media files')

    objects = TranslatedObjectManager()

    filetypes = []
    filetypes_dict = {}

    @classmethod
    def reconfigure(cls, upload_to=None, storage=None):
        f = cls._meta.get_field('file')
        # Ugh. Copied relevant parts from django/db/models/fields/files.py
        # FileField.__init__ (around line 225)
        if storage:
            f.storage = storage
        if upload_to:
            f.upload_to = upload_to
            if callable(upload_to):
                f.generate_filename = upload_to

    @classmethod
    def register_filetypes(cls, *types):
        cls.filetypes[0:0] = types
        choices = [t[0:2] for t in cls.filetypes]
        cls.filetypes_dict = dict(choices)
        cls._meta.get_field('type').choices[:] = choices

    def __init__(self, *args, **kwargs):
        super(MediaFileBase, self).__init__(*args, **kwargs)
        if self.file:
            self._original_file_name = self.file.name

    def __str__(self):
        trans = None

        try:
            trans = self.translation
        except models.ObjectDoesNotExist:
            pass
        except AttributeError:
            pass

        if trans:
            trans = '%s' % trans
            if trans.strip():
                return trans
        return os.path.basename(self.file.name)

    def get_absolute_url(self):
        return self.file.url

    def determine_file_type(self, name):
        """
        >>> t = MediaFileBase()
        >>> str(t.determine_file_type('foobar.jpg'))
        'image'
        >>> str(t.determine_file_type('foobar.PDF'))
        'pdf'
        >>> str(t.determine_file_type('foobar.jpg.pdf'))
        'pdf'
        >>> str(t.determine_file_type('foobar.jgp'))
        'other'
        >>> str(t.determine_file_type('foobar-jpg'))
        'other'
        """
        for type_key, type_name, type_test in self.filetypes:
            if type_test(name):
                return type_key
        return self.filetypes[-1][0]

    def save(self, *args, **kwargs):
        if not self.id and not self.created:
            self.created = timezone.now()

        self.type = self.determine_file_type(self.file.name)
        if self.file:
            try:
                self.file_size = self.file.size
            except (OSError, IOError, ValueError) as e:
                logger.error("Unable to read file size for %s: %s" % (self, e))

        super(MediaFileBase, self).save(*args, **kwargs)

        logger.info("Saved mediafile %d (%s, type %s, %d bytes)" % (
            self.id, self.file.name, self.type, self.file_size or 0))

        # User uploaded a new file. Try to get rid of the old file in
        # storage, to avoid having orphaned files hanging around.
        if getattr(self, '_original_file_name', None):
            if self.file.name != self._original_file_name:
                self.delete_mediafile(self._original_file_name)

        self.purge_translation_cache()
    save.alters_data = True

    def delete_mediafile(self, name=None):
        if name is None:
            name = self.file.name
        try:
            self.file.storage.delete(name)
        except Exception as e:
            logger.warn("Cannot delete media file %s: %s" % (name, e))


# ------------------------------------------------------------------------
MediaFileBase.register_filetypes(
    # Should we be using imghdr.what instead of extension guessing?
    ('image', _('Image'), lambda f: re.compile(
        r'\.(bmp|jpe?g|jp2|jxr|gif|png|tiff?)$', re.IGNORECASE).search(f)),
    ('video', _('Video'), lambda f: re.compile(
        r'\.(mov|m[14]v|mp4|avi|mpe?g|qt|ogv|wmv|flv)$',
        re.IGNORECASE).search(f)),
    ('audio', _('Audio'), lambda f: re.compile(
        r'\.(au|mp3|m4a|wma|oga|ram|wav)$', re.IGNORECASE).search(f)),
    ('pdf', _('PDF document'), lambda f: f.lower().endswith('.pdf')),
    ('swf', _('Flash'), lambda f: f.lower().endswith('.swf')),
    ('txt', _('Text'), lambda f: f.lower().endswith('.txt')),
    ('rtf', _('Rich Text'), lambda f: f.lower().endswith('.rtf')),
    ('zip', _('Zip archive'), lambda f: f.lower().endswith('.zip')),
    ('doc', _('Microsoft Word'), lambda f: re.compile(
        r'\.docx?$', re.IGNORECASE).search(f)),
    ('xls', _('Microsoft Excel'), lambda f: re.compile(
        r'\.xlsx?$', re.IGNORECASE).search(f)),
    ('ppt', _('Microsoft PowerPoint'), lambda f: re.compile(
        r'\.pptx?$', re.IGNORECASE).search(f)),
    ('other', _('Binary'), lambda f: True),  # Must be last
)


# ------------------------------------------------------------------------
class MediaFile(MediaFileBase):
    class Meta:
        app_label = 'medialibrary'


@receiver(post_delete, sender=MediaFile)
def _mediafile_post_delete(sender, instance, **kwargs):
    instance.delete_mediafile()
    logger.info("Deleted mediafile %d (%s)" % (
        instance.id, instance.file.name))


# ------------------------------------------------------------------------
@python_2_unicode_compatible
class MediaFileTranslation(Translation(MediaFile)):
    """
    Translated media file caption and description.
    """

    caption = models.CharField(_('caption'), max_length=1000)
    description = models.TextField(_('description'), blank=True)

    class Meta:
        verbose_name = _('media file translation')
        verbose_name_plural = _('media file translations')
        unique_together = ('parent', 'language_code')
        app_label = 'medialibrary'

    def __str__(self):
        return self.caption



from __future__ import absolute_import, unicode_literals

import re

from django.db import models
from django.utils.translation import ugettext_lazy as _

from feincms._internal import ct_render_to_string


class VideoContent(models.Model):
    """
    Copy-paste a URL to youtube or vimeo into the text box, this content type
    will automatically generate the necessary embed code.

    Other portals aren't supported currently, but would be easy to add if
    anyone would take up the baton.

    You should probably use feincms-oembed.
    """

    PORTALS = (
        ('youtube', re.compile(r'youtube'), lambda url: {
            'v': re.search(r'([?&]v=|./././)([^#&]+)', url).group(2),
        }),
        ('vimeo', re.compile(r'vimeo'), lambda url: {
            'id': re.search(r'/(\d+)', url).group(1),
        }),
        ('sf', re.compile(r'sf\.tv'), lambda url: {
            'id': re.search(r'/([a-z0-9\-]+)', url).group(1),
        }),
    )

    video = models.URLField(
        _('video link'),
        help_text=_(
            'This should be a link to a youtube or vimeo video,'
            ' i.e.: http://www.youtube.com/watch?v=zmj1rpzDRZ0'))

    class Meta:
        abstract = True
        verbose_name = _('video')
        verbose_name_plural = _('videos')

    def get_context_dict(self):
        "Extend this if you need more variables passed to template"
        return {'content': self, 'portal': 'unknown'}

    def get_templates(self, portal='unknown'):
        "Extend/override this if you want to modify the templates used"
        return [
            'content/video/%s.html' % portal,
            'content/video/unknown.html',
        ]

    def ctx_for_video(self, vurl):
        "Get a context dict for a given video URL"
        ctx = self.get_context_dict()
        for portal, match, context_fn in self.PORTALS:
            if match.search(vurl):
                try:
                    ctx.update(context_fn(vurl))
                    ctx['portal'] = portal
                    break
                except AttributeError:
                    continue
        return ctx

    def render(self, **kwargs):
        ctx = self.ctx_for_video(self.video)
        return ct_render_to_string(
            self.get_templates(ctx['portal']),
            ctx,
            request=kwargs.get('request'),
            context=kwargs.get('context'),
        )



from __future__ import absolute_import, unicode_literals

import os

from django.core.management.base import NoArgsCommand
from django.utils.encoding import force_text

from feincms.module.medialibrary.models import MediaFile


class Command(NoArgsCommand):
    help = "Prints all orphaned files in the `media/medialibrary` folder"

    def handle_noargs(self, **options):
        mediafiles = list(MediaFile.objects.values_list('file', flat=True))

        # TODO make this smarter, and take MEDIA_ROOT into account
        for base, dirs, files in os.walk('media/medialibrary'):
            for f in files:
                full = os.path.join(base[6:], f)
                if force_text(full) not in mediafiles:
                    self.stdout.write(os.path.join(base, f))



"""
Add an excerpt field to the page.
"""

from __future__ import absolute_import, unicode_literals

from django.db import models
from django.utils.translation import ugettext_lazy as _

from feincms import extensions


class Extension(extensions.Extension):
    def handle_model(self):
        self.model.add_to_class(
            'excerpt',
            models.TextField(
                _('excerpt'),
                blank=True,
                help_text=_(
                    'Add a brief excerpt summarizing the content'
                    ' of this page.')))

    def handle_modeladmin(self, modeladmin):
        modeladmin.add_extension_options(_('Excerpt'), {
            'fields': ('excerpt',),
            'classes': ('collapse',),
        })



from django.conf.urls import url

from feincms.contrib.preview.views import PreviewHandler


urlpatterns = [
    url(r'^(.*)/_preview/(\d+)/$', PreviewHandler.as_view(),
        name='feincms_preview'),
]



# ------------------------------------------------------------------------
# coding=utf-8
# ------------------------------------------------------------------------

"""
This extension adds a language field to every page. When calling the request
processors the page's language is activated.
Pages in secondary languages can be said to be a translation of a page in the
primary language (the first language in settings.LANGUAGES), thereby enabling
deeplinks between translated pages.

It is recommended to activate
:class:`django.middleware.locale.LocaleMiddleware` so that the correct language
will be activated per user or session even for non-FeinCMS managed views such
as Django's administration tool.
"""

from __future__ import absolute_import, unicode_literals

# ------------------------------------------------------------------------
import logging

from django.conf import settings as django_settings
from django.db import models
from django.http import HttpResponseRedirect
from django.utils import translation
from django.utils.html import mark_safe
from django.utils.translation import ugettext_lazy as _

from feincms import extensions, settings
from feincms.translations import is_primary_language
from feincms._internal import monkeypatch_method, monkeypatch_property


# ------------------------------------------------------------------------
logger = logging.getLogger(__name__)

LANGUAGE_COOKIE_NAME = django_settings.LANGUAGE_COOKIE_NAME
if hasattr(translation, 'LANGUAGE_SESSION_KEY'):
    LANGUAGE_SESSION_KEY = translation.LANGUAGE_SESSION_KEY
else:
    # Django 1.6
    LANGUAGE_SESSION_KEY = LANGUAGE_COOKIE_NAME


# ------------------------------------------------------------------------
def user_has_language_set(request):
    """
    Determine whether the user has explicitely set a language earlier on.
    This is taken later on as an indication that we should not mess with the
    site's language settings, after all, the user's decision is what counts.
    """
    if (hasattr(request, 'session') and
            request.session.get(LANGUAGE_SESSION_KEY) is not None):
        return True
    if LANGUAGE_COOKIE_NAME in request.COOKIES:
        return True
    return False


# ------------------------------------------------------------------------
def translation_allowed_language(select_language):
    "Check for feincms specific set of allowed front end languages."
    if settings.FEINCMS_FRONTEND_LANGUAGES:
        language = select_language[:2]
        if language not in settings.FEINCMS_FRONTEND_LANGUAGES:
            select_language = django_settings.LANGUAGES[0][0]

    return select_language


# ------------------------------------------------------------------------
def translation_set_language(request, select_language):
    """
    Set and activate a language, if that language is available.
    """

    select_language = translation_allowed_language(select_language)

    if translation.check_for_language(select_language):
        fallback = False
    else:
        # The page is in a language that Django has no messages for.
        # We display anyhow, but fall back to primary language for
        # other messages and other applications. It is *highly* recommended to
        # create a new django.po for the language instead of
        # using this behaviour.
        select_language = django_settings.LANGUAGES[0][0]
        fallback = True

    translation.activate(select_language)
    request.LANGUAGE_CODE = translation.get_language()

    if hasattr(request, 'session'):
        # User has a session, then set this language there
        if select_language != request.session.get(LANGUAGE_SESSION_KEY):
            request.session[LANGUAGE_SESSION_KEY] = select_language
    elif request.method == 'GET' and not fallback:
        # No session is active. We need to set a cookie for the language
        # so that it persists when users change their location to somewhere
        # not under the control of the CMS.
        # Only do this when request method is GET (mainly, do not abort
        # POST requests)
        response = HttpResponseRedirect(request.get_full_path())
        response.set_cookie(
            str(LANGUAGE_COOKIE_NAME), select_language)
        return response


# ------------------------------------------------------------------------
def translations_request_processor_explicit(page, request):
    # If this page is just a redirect, don't do any language specific setup
    if page.redirect_to:
        return

    # Until further notice, the user might be wanting to switch to the
    # page's language...
    desired_language = page.language

    # ...except if the user explicitely wants to switch language
    if 'set_language' in request.GET:
        desired_language = request.GET['set_language']
    # ...or the user already has explicitely set a language, bail out and
    # don't change it for them behind their back
    elif user_has_language_set(request):
        return

    return translation_set_language(request, desired_language)


# ------------------------------------------------------------------------
def translations_request_processor_standard(page, request):
    # If this page is just a redirect, don't do any language specific setup
    if getattr(page, 'redirect_to', None):
        return

    if page.language == translation.get_language():
        return

    return translation_set_language(request, page.language)


# ------------------------------------------------------------------------
def get_current_language_code(request):
    language_code = getattr(request, 'LANGUAGE_CODE', None)
    if language_code is None:
        logger.warning(
            "Could not access request.LANGUAGE_CODE. Is 'django.middleware."
            "locale.LocaleMiddleware' in MIDDLEWARE_CLASSES?")
    return language_code


# ------------------------------------------------------------------------
class Extension(extensions.Extension):

    def handle_model(self):
        cls = self.model

        cls.add_to_class(
            'language',
            models.CharField(
                _('language'),
                max_length=10,
                choices=django_settings.LANGUAGES,
                default=django_settings.LANGUAGES[0][0]))
        cls.add_to_class(
            'translation_of',
            models.ForeignKey(
                'self',
                on_delete=models.CASCADE,
                blank=True, null=True, verbose_name=_('translation of'),
                related_name='translations',
                limit_choices_to={'language': django_settings.LANGUAGES[0][0]},
                help_text=_(
                    'Leave this empty for entries in the primary language.'),
            )
        )

        if hasattr(cls, 'register_request_processor'):
            if settings.FEINCMS_TRANSLATION_POLICY == "EXPLICIT":
                cls.register_request_processor(
                    translations_request_processor_explicit,
                    key='translations')
            else:  # STANDARD
                cls.register_request_processor(
                    translations_request_processor_standard,
                    key='translations')

        if hasattr(cls, 'get_redirect_to_target'):
            original_get_redirect_to_target = cls.get_redirect_to_target

            @monkeypatch_method(cls)
            def get_redirect_to_target(self, request=None):
                """
                Find an acceptable redirect target. If this is a local link,
                then try to find the page this redirect references and
                translate it according to the user's language. This way, one
                can easily implement a localized "/"-url to welcome page
                redirection.
                """
                target = original_get_redirect_to_target(self, request)
                if target and target.find('//') == -1:
                    # Not an offsite link http://bla/blubb
                    try:
                        page = cls.objects.page_for_path(target)
                        language = get_current_language_code(request)
                        language = translation_allowed_language(language)
                        page = page.get_translation(language)
                        # Note: Does not care about active status?
                        target = page.get_absolute_url()
                    except cls.DoesNotExist:
                        pass
                return target

        @monkeypatch_method(cls)
        def available_translations(self):
            if not self.id:  # New, unsaved pages have no translations
                return []

            if hasattr(cls.objects, 'apply_active_filters'):
                filter_active = cls.objects.apply_active_filters
            else:
                def filter_active(queryset):
                    return queryset

            if is_primary_language(self.language):
                return filter_active(self.translations.all())
            elif self.translation_of:
                # reuse prefetched queryset, do not filter it
                res = [
                    t for t
                    in filter_active(self.translation_of.translations.all())
                    if t.language != self.language]
                res.insert(0, self.translation_of)
                return res
            else:
                return []

        @monkeypatch_method(cls)
        def get_original_translation(self, *args, **kwargs):
            if is_primary_language(self.language):
                return self
            if self.translation_of:
                return self.translation_of
            logger.debug(
                "Page pk=%d (%s) has no primary language translation (%s)",
                self.pk, self.language, django_settings.LANGUAGES[0][0])
            return self

        @monkeypatch_property(cls)
        def original_translation(self):
            return self.get_original_translation()

        @monkeypatch_method(cls)
        def get_translation(self, language):
            return self.original_translation.translations.get(
                language=language)

    def handle_modeladmin(self, modeladmin):
        extensions.prefetch_modeladmin_get_queryset(
            modeladmin, 'translation_of__translations', 'translations')

        def available_translations_admin(self, page):
            # Do not use available_translations() because we don't care
            # whether pages are active or not here.
            translations = [page]
            translations.extend(page.translations.all())
            if page.translation_of:
                translations.append(page.translation_of)
                translations.extend(page.translation_of.translations.all())
            translations = {
                p.language: p.id
                for p in translations
            }

            links = []

            for key, title in django_settings.LANGUAGES:
                if key == page.language:
                    continue

                if key in translations:
                    links.append('<a href="%s/" title="%s">%s</a>' % (
                        translations[key], _('Edit translation'), key.upper()))
                else:
                    links.append(
                        '<a style="color:#baa" href="add/?translation_of='
                        '%s&amp;language=%s" title="%s">%s</a>' % (
                            page.id,
                            key,
                            _('Create translation'),
                            key.upper()
                        )
                    )

            return mark_safe(' | '.join(links))

        available_translations_admin.short_description = _('translations')
        modeladmin.__class__.available_translations_admin =\
            available_translations_admin

        if hasattr(modeladmin, 'add_extension_options'):
            modeladmin.add_extension_options('language', 'translation_of')

        modeladmin.extend_list(
            'list_display',
            ['language', 'available_translations_admin'],
        )
        modeladmin.extend_list('list_filter', ['language'])
        modeladmin.extend_list('raw_id_fields', ['translation_of'])

# ------------------------------------------------------------------------



from __future__ import absolute_import, unicode_literals

from django.contrib import admin
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.translation import ugettext_lazy as _

from feincms.admin.item_editor import FeinCMSInline
from feincms._internal import ct_render_to_string

try:
    from filer.fields.file import FilerFileField
    from filer.fields.image import FilerImageField
except ImportError:
    __all__ = ()

else:

    __all__ = (
        'MediaFileContentInline', 'ContentWithFilerFile',
        'FilerFileContent', 'FilerImageContent',
    )

    class MediaFileContentInline(FeinCMSInline):
        radio_fields = {'type': admin.VERTICAL}

    class ContentWithFilerFile(models.Model):
        """
        File content
        """
        feincms_item_editor_inline = MediaFileContentInline

        class Meta:
            abstract = True

        def render(self, **kwargs):
            return ct_render_to_string(
                [
                    'content/filer/%s_%s.html' % (self.file_type, self.type),
                    'content/filer/%s.html' % self.type,
                    'content/filer/%s.html' % self.file_type,
                    'content/filer/default.html',
                ],
                {'content': self},
                request=kwargs.get('request'),
                context=kwargs.get('context'),
            )

    class FilerFileContent(ContentWithFilerFile):
        mediafile = FilerFileField(verbose_name=_('file'), related_name='+')
        file_type = 'file'
        type = 'download'

        class Meta:
            abstract = True
            verbose_name = _('file')
            verbose_name_plural = _('files')

    class FilerImageContent(ContentWithFilerFile):
        """
        Create a media file content as follows::

            from feincms.contents import FilerImageContent
            Page.create_content_type(FilerImageContent, TYPE_CHOICES=(
                ('inline', _('Default')),
                ('lightbox', _('Lightbox')),
                ('whatever', _('Whatever')),
            ))

        For a media file of type 'image' and type 'lightbox', the following
        templates are tried in order:

        * content/mediafile/image_lightbox.html
        * content/mediafile/lightbox.html
        * content/mediafile/image.html
        * content/mediafile/default.html

        The context contains ``content`` and ``request`` (if available).

        The content.mediafile attribute are as follows (selection):
        label, description, default_caption, default_alt_text,
        author, must_always_publish_author_credit,
        must_always_publish_copyright, date_taken, file, id, is_public, url
        """

        mediafile = FilerImageField(verbose_name=_('image'), related_name='+')
        caption = models.CharField(
            _('caption'),
            max_length=1000,
            blank=True,
        )
        url = models.CharField(
            _('URL'),
            max_length=1000,
            blank=True,
        )

        file_type = 'image'

        class Meta:
            abstract = True
            verbose_name = _('image')
            verbose_name_plural = _('images')

        @classmethod
        def initialize_type(cls, TYPE_CHOICES=None):
            if TYPE_CHOICES is None:
                raise ImproperlyConfigured(
                    'You have to set TYPE_CHOICES when'
                    ' creating a %s' % cls.__name__)

            cls.add_to_class(
                'type',
                models.CharField(
                    _('type'),
                    max_length=20,
                    choices=TYPE_CHOICES,
                    default=TYPE_CHOICES[0][0],
                ),
            )



from __future__ import absolute_import, unicode_literals

from feincms import settings
from feincms.templatetags import feincms_thumbnail
from feincms.utils import get_object


def default_admin_thumbnail(mediafile, dimensions='100x100', **kwargs):
    if mediafile.type != 'image':
        return None

    return feincms_thumbnail.thumbnail(mediafile.file, dimensions)


_cached_thumbnailer = None


def admin_thumbnail(mediafile, dimensions='100x100'):
    global _cached_thumbnailer
    if not _cached_thumbnailer:
        _cached_thumbnailer = get_object(
            settings.FEINCMS_MEDIALIBRARY_THUMBNAIL)
    return _cached_thumbnailer(mediafile, dimensions=dimensions)



'''
I really hate repeating myself. These are helpers that avoid typing the
whole thing over and over when implementing additional template tags

They help implementing tags of the following forms::

    {% tag as var_name %}
    {% tag of template_var as var_name %}
    {% tag of template_var as var_name arg1,arg2,kwarg3=4 %}
'''

from __future__ import absolute_import, unicode_literals

from django import template


def _parse_args(argstr, context=None):
    try:
        args = {}
        for token in argstr.split(','):
            if '=' in token:
                k, v = token.split('=', 1)
                if context:
                    try:
                        args[k] = template.Variable(v).resolve(context)
                    except template.VariableDoesNotExist:
                        args[k] = v
                else:
                    args[k] = v
            else:
                args[token] = True

        return args

    except TypeError:
        raise template.TemplateSyntaxError('Malformed arguments')


def do_simple_assignment_node_with_var_and_args_helper(cls):
    def _func(parser, token):
        try:
            tag_name, of_, in_var_name, as_, var_name, args =\
                token.contents.split()
        except ValueError:
            try:
                tag_name, of_, in_var_name, as_, var_name =\
                    token.contents.split()
                args = ''
            except ValueError:
                raise template.TemplateSyntaxError(
                    'Invalid syntax for %s node: %s' % (
                        cls.__name__, token.contents))

        return cls(tag_name, in_var_name, var_name, args)

    return _func


class SimpleAssignmentNodeWithVarAndArgs(template.Node):
    def __init__(self, tag_name, in_var_name, var_name, args):
        self.tag_name = tag_name
        self.in_var = template.Variable(in_var_name)
        self.var_name = var_name
        self.args = args

    def render(self, context):
        self.render_context = context
        try:
            instance = self.in_var.resolve(context)
        except template.VariableDoesNotExist:
            context[self.var_name] = []
            return ''

        context[self.var_name] = self.what(
            instance, _parse_args(self.args, context))

        return ''



# flake8: noqa
from __future__ import absolute_import, unicode_literals

import warnings

from feincms.module.medialibrary.contents import MediaFileContent

warnings.warn(
    'Import MediaFileContent from feincms.module.medialibrary.contents.',
    DeprecationWarning, stacklevel=2)



# ------------------------------------------------------------------------
# coding=utf-8
# ------------------------------------------------------------------------
"""
Default settings for FeinCMS

All of these can be overridden by specifying them in the standard
``settings.py`` file.
"""

from __future__ import absolute_import, unicode_literals

from django.conf import settings

# e.g. 'uploads' if you would prefer <media root>/uploads/imagecontent/test.jpg
# to <media root>/imagecontent/test.jpg.
FEINCMS_UPLOAD_PREFIX = getattr(
    settings,
    'FEINCMS_UPLOAD_PREFIX',
    '')

# ------------------------------------------------------------------------
# Settings for MediaLibrary

#: Local path to newly uploaded media files
FEINCMS_MEDIALIBRARY_UPLOAD_TO = getattr(
    settings,
    'FEINCMS_MEDIALIBRARY_UPLOAD_TO',
    'medialibrary/%Y/%m/')

#: Thumbnail function for suitable mediafiles. Only receives the media file
#: and should return a thumbnail URL (or nothing).
FEINCMS_MEDIALIBRARY_THUMBNAIL = getattr(
    settings,
    'FEINCMS_MEDIALIBRARY_THUMBNAIL',
    'feincms.module.medialibrary.thumbnail.default_admin_thumbnail')

# ------------------------------------------------------------------------
# Settings for RichText

FEINCMS_RICHTEXT_INIT_TEMPLATE = getattr(
    settings,
    'FEINCMS_RICHTEXT_INIT_TEMPLATE',
    'admin/content/richtext/init_tinymce4.html')
FEINCMS_RICHTEXT_INIT_CONTEXT = getattr(
    settings,
    'FEINCMS_RICHTEXT_INIT_CONTEXT', {
        'TINYMCE_JS_URL': '//tinymce.cachefly.net/4.2/tinymce.min.js',
        'TINYMCE_DOMAIN': None,
        'TINYMCE_CONTENT_CSS_URL': None,
        'TINYMCE_LINK_LIST_URL': None
    }
)

# ------------------------------------------------------------------------
# Settings for the page module

#: Include ancestors in filtered tree editor lists
FEINCMS_TREE_EDITOR_INCLUDE_ANCESTORS = getattr(
    settings,
    'FEINCMS_TREE_EDITOR_INCLUDE_ANCESTORS',
    False)

#: Enable checking of object level permissions. Note that if this option is
#: enabled, you must plug in an authentication backend that actually does
#: implement object level permissions or no page will be editable.
FEINCMS_TREE_EDITOR_OBJECT_PERMISSIONS = getattr(
    settings,
    'FEINCMS_TREE_EDITOR_OBJECT_PERMISSIONS',
    False)

#: When enabled, the page module is automatically registered with Django's
#: default admin site (this is activated by default).
FEINCMS_USE_PAGE_ADMIN = getattr(
    settings,
    'FEINCMS_USE_PAGE_ADMIN',
    True)

#: app_label.model_name as per apps.get_model.
#: defaults to page.Page
FEINCMS_DEFAULT_PAGE_MODEL = getattr(
    settings,
    'FEINCMS_DEFAULT_PAGE_MODEL',
    'page.Page')

# ------------------------------------------------------------------------
#: Allow random gunk after a valid page?
FEINCMS_ALLOW_EXTRA_PATH = getattr(
    settings,
    'FEINCMS_ALLOW_EXTRA_PATH',
    False)

# ------------------------------------------------------------------------
#: How to switch languages.
#: * ``'STANDARD'``: The page a user navigates to sets the site's language
#:   and overwrites whatever was set before.
#: * ``'EXPLICIT'``: The language set has priority, may only be overridden
#:   by explicitely a language with ``?set_language=xx``.
FEINCMS_TRANSLATION_POLICY = getattr(
    settings,
    'FEINCMS_TRANSLATION_POLICY',
    'STANDARD')

# ------------------------------------------------------------------------
#: Makes the page handling mechanism try to find a cms page with that
#: path if it encounters a page not found situation. This allows for nice
#: customised cms-styled error pages. Do not go overboard, this should
#: be as simple and as error resistant as possible, so refrain from
#: deeply nested error pages or advanced content types.
FEINCMS_CMS_404_PAGE = getattr(
    settings,
    'FEINCMS_CMS_404_PAGE',
    None)

# ------------------------------------------------------------------------
#: When uploading files to the media library, replacing an existing entry,
#: try to save the new file under the old file name in order to keep the
#: media file path (and thus the media url) constant.
#: Experimental, this might not work with all storage backends.
FEINCMS_MEDIAFILE_OVERWRITE = getattr(
    settings,
    'FEINCMS_MEDIAFILE_OVERWRITE',
    False)

# ------------------------------------------------------------------------
#: Prefix for thumbnails. Set this to something non-empty to separate thumbs
#: from uploads. The value should end with a slash, but this is not enforced.
FEINCMS_THUMBNAIL_DIR = getattr(
    settings,
    'FEINCMS_THUMBNAIL_DIR',
    '_thumbs/')

# ------------------------------------------------------------------------
#: feincms_thumbnail template filter library cache timeout. The default is to
#: not cache anything for backwards compatibility.
FEINCMS_THUMBNAIL_CACHE_TIMEOUT = getattr(
    settings,
    'FEINCMS_THUMBNAIL_CACHE_TIMEOUT',
    0)

# ------------------------------------------------------------------------
#: Prevent changing template within admin for pages which have been
#: allocated a Template with singleton=True -- template field will become
#: read-only for singleton pages.
FEINCMS_SINGLETON_TEMPLATE_CHANGE_ALLOWED = getattr(
    settings,
    'FEINCMS_SINGLETON_TEMPLATE_CHANGE_ALLOWED',
    False)

#: Prevent admin page deletion for pages which have been allocated a
#: Template with singleton=True
FEINCMS_SINGLETON_TEMPLATE_DELETION_ALLOWED = getattr(
    settings,
    'FEINCMS_SINGLETON_TEMPLATE_DELETION_ALLOWED',
    False)

# ------------------------------------------------------------------------
#: Filter languages available for front end users to this set. This allows
#: to have languages not yet ready for prime time while being able to access
#: those pages in the admin backend.
FEINCMS_FRONTEND_LANGUAGES = getattr(
    settings,
    'FEINCMS_FRONTEND_LANGUAGES',
    None)

# ------------------------------------------------------------------------



from __future__ import absolute_import

from collections import OrderedDict
from email.utils import parsedate
from functools import partial, wraps
from time import mktime
import warnings

from django.conf import settings
from django.core.cache import cache
from django.db import models
from django.http import HttpResponse
from django.template.response import TemplateResponse
from django.utils.functional import lazy
from django.utils.http import http_date
from django.utils.safestring import mark_safe
from django.utils.translation import get_language, ugettext_lazy as _
try:
    from django.urls import (
        NoReverseMatch, reverse, get_script_prefix, set_script_prefix,
        Resolver404, resolve,
    )
except ImportError:
    from django.core.urlresolvers import (
        NoReverseMatch, reverse, get_script_prefix, set_script_prefix,
        Resolver404, resolve,
    )

from feincms.admin.item_editor import ItemEditorForm
from feincms.contrib.fields import JSONField
from feincms.translations import short_language_code
from feincms.utils import get_object


APP_REVERSE_CACHE_TIMEOUT = 3


__all__ = (
    'ApplicationContent',
    'app_reverse', 'app_reverse_lazy', 'permalink',
    'UnpackTemplateResponse', 'standalone', 'unpack',
)


class UnpackTemplateResponse(TemplateResponse):
    """
    Completely the same as marking applicationcontent-contained views with
    the ``feincms.views.decorators.unpack`` decorator.
    """
    _feincms_unpack = True


def standalone(view_func):
    """
    Marks the view method as standalone view; this means that
    ``HttpResponse`` objects returned from ``ApplicationContent``
    are returned directly, without further processing.
    """

    def inner(request, *args, **kwargs):
        response = view_func(request, *args, **kwargs)
        if isinstance(response, HttpResponse):
            response.standalone = True
        return response
    return wraps(view_func)(inner)


def unpack(view_func):
    """
    Marks the returned response as to-be-unpacked if it is a
    ``TemplateResponse``.
    """

    def inner(request, *args, **kwargs):
        response = view_func(request, *args, **kwargs)
        if isinstance(response, TemplateResponse):
            response._feincms_unpack = True
        return response
    return wraps(view_func)(inner)


def cycle_app_reverse_cache(*args, **kwargs):
    warnings.warn(
        'cycle_app_reverse_cache does nothing and will be removed in'
        ' a future version of FeinCMS.',
        DeprecationWarning, stacklevel=2,
    )


def app_reverse(viewname, urlconf=None, args=None, kwargs=None,
                *vargs, **vkwargs):
    """
    Reverse URLs from application contents

    Works almost like Django's own reverse() method except that it resolves
    URLs from application contents. The second argument, ``urlconf``, has to
    correspond to the URLconf parameter passed in the ``APPLICATIONS`` list
    to ``Page.create_content_type``::

        app_reverse('mymodel-detail', 'myapp.urls', args=...)

        or

        app_reverse('mymodel-detail', 'myapp.urls', kwargs=...)

    The second argument may also be a request object if you want to reverse
    an URL belonging to the current application content.
    """

    # First parameter might be a request instead of an urlconf path, so
    # we'll try to be helpful and extract the current urlconf from it
    extra_context = getattr(urlconf, '_feincms_extra_context', {})
    appconfig = extra_context.get('app_config', {})
    urlconf = appconfig.get('urlconf_path', urlconf)

    appcontent_class = ApplicationContent._feincms_content_models[0]
    cache_key = appcontent_class.app_reverse_cache_key(urlconf)
    url_prefix = cache.get(cache_key)

    if url_prefix is None:
        content = appcontent_class.closest_match(urlconf)

        if content is not None:
            if urlconf in appcontent_class.ALL_APPS_CONFIG:
                # We have an overridden URLconf
                app_config = appcontent_class.ALL_APPS_CONFIG[urlconf]
                urlconf = app_config['config'].get('urls', urlconf)

            prefix = content.parent.get_absolute_url()
            prefix += '/' if prefix[-1] != '/' else ''

            url_prefix = (urlconf, prefix)
            cache.set(cache_key, url_prefix, timeout=APP_REVERSE_CACHE_TIMEOUT)

    if url_prefix:
        # vargs and vkwargs are used to send through additional parameters
        # which are uninteresting to us (such as current_app)
        prefix = get_script_prefix()
        try:
            set_script_prefix(url_prefix[1])
            return reverse(
                viewname,
                url_prefix[0],
                args=args,
                kwargs=kwargs,
                *vargs, **vkwargs)
        finally:
            set_script_prefix(prefix)

    raise NoReverseMatch("Unable to find ApplicationContent for %r" % urlconf)


#: Lazy version of ``app_reverse``
app_reverse_lazy = lazy(app_reverse, str)


def permalink(func):
    """
    Decorator that calls app_reverse()

    Use this instead of standard django.db.models.permalink if you want to
    integrate the model through ApplicationContent. The wrapped function
    must return 4 instead of 3 arguments::

        class MyModel(models.Model):
            @appmodels.permalink
            def get_absolute_url(self):
                return ('myapp.urls', 'model_detail', (), {'slug': self.slug})
    """
    def inner(*args, **kwargs):
        return app_reverse(*func(*args, **kwargs))
    return wraps(func)(inner)


class ApplicationContent(models.Model):
    #: parameters is used to serialize instance-specific data which will be
    # provided to the view code. This allows customization (e.g. "Embed
    # MyBlogApp for blog <slug>")
    parameters = JSONField(null=True, editable=False)

    ALL_APPS_CONFIG = OrderedDict()

    class Meta:
        abstract = True
        verbose_name = _('application content')
        verbose_name_plural = _('application contents')

    @classmethod
    def initialize_type(cls, APPLICATIONS):
        for i in APPLICATIONS:
            if not 2 <= len(i) <= 3:
                raise ValueError(
                    "APPLICATIONS must be provided with tuples containing at"
                    " least two parameters (urls, name) and an optional extra"
                    " config dict")

            urls, name = i[0:2]

            if len(i) == 3:
                app_conf = i[2]

                if not isinstance(app_conf, dict):
                    raise ValueError(
                        "The third parameter of an APPLICATIONS entry must be"
                        " a dict or the name of one!")
            else:
                app_conf = {}

            cls.ALL_APPS_CONFIG[urls] = {
                "urls": urls,
                "name": name,
                "config": app_conf
            }

        cls.add_to_class(
            'urlconf_path',
            models.CharField(_('application'), max_length=100, choices=[
                (c['urls'], c['name']) for c in cls.ALL_APPS_CONFIG.values()])
        )

        class ApplicationContentItemEditorForm(ItemEditorForm):
            app_config = {}
            custom_fields = {}

            def __init__(self, *args, **kwargs):
                super(ApplicationContentItemEditorForm, self).__init__(
                    *args, **kwargs)

                instance = kwargs.get("instance", None)

                if instance:
                    try:
                        # TODO use urlconf_path from POST if set
                        # urlconf_path = request.POST.get('...urlconf_path',
                        #     instance.urlconf_path)
                        self.app_config = cls.ALL_APPS_CONFIG[
                            instance.urlconf_path]['config']
                    except KeyError:
                        self.app_config = {}

                    self.custom_fields = {}
                    admin_fields = self.app_config.get('admin_fields', {})

                    if isinstance(admin_fields, dict):
                        self.custom_fields.update(admin_fields)
                    else:
                        get_fields = get_object(admin_fields)
                        self.custom_fields.update(
                            get_fields(self, *args, **kwargs))

                    params = self.instance.parameters
                    for k, v in self.custom_fields.items():
                        v.initial = params.get(k)
                        self.fields[k] = v
                        if k in params:
                            self.fields[k].initial = params[k]

            def save(self, commit=True, *args, **kwargs):
                # Django ModelForms return the model instance from save. We'll
                # call save with commit=False first to do any necessary work &
                # get the model so we can set .parameters to the values of our
                # custom fields before calling save(commit=True)

                m = super(ApplicationContentItemEditorForm, self).save(
                    commit=False, *args, **kwargs)

                m.parameters = dict(
                    (k, self.cleaned_data[k])
                    for k in self.custom_fields if k in self.cleaned_data)

                if commit:
                    m.save(**kwargs)

                return m

        # This provides hooks for us to customize the admin interface for
        # embedded instances:
        cls.feincms_item_editor_form = ApplicationContentItemEditorForm

    def __init__(self, *args, **kwargs):
        super(ApplicationContent, self).__init__(*args, **kwargs)
        self.app_config = self.ALL_APPS_CONFIG.get(
            self.urlconf_path, {}).get('config', {})

    def process(self, request, **kw):
        page_url = self.parent.get_absolute_url()

        # Provide a way for appcontent items to customize URL processing by
        # altering the perceived path of the page:
        if "path_mapper" in self.app_config:
            path_mapper = get_object(self.app_config["path_mapper"])
            path, page_url = path_mapper(
                request.path,
                page_url,
                appcontent_parameters=self.parameters
            )
        else:
            path = request._feincms_extra_context['extra_path']

        # Resolve the module holding the application urls.
        urlconf_path = self.app_config.get('urls', self.urlconf_path)

        try:
            fn, args, kwargs = resolve(path, urlconf_path)
        except (ValueError, Resolver404):
            raise Resolver404(str('Not found (resolving %r in %r failed)') % (
                path, urlconf_path))

        # Variables from the ApplicationContent parameters are added to request
        # so we can expose them to our templates via the appcontent_parameters
        # context_processor
        request._feincms_extra_context.update(self.parameters)

        # Save the application configuration for reuse elsewhere
        request._feincms_extra_context.update({
            'app_config': dict(
                self.app_config,
                urlconf_path=self.urlconf_path,
            ),
        })

        view_wrapper = self.app_config.get("view_wrapper", None)
        if view_wrapper:
            fn = partial(
                get_object(view_wrapper),
                view=fn,
                appcontent_parameters=self.parameters
            )

        output = fn(request, *args, **kwargs)

        if isinstance(output, HttpResponse):
            if self.send_directly(request, output):
                return output
            elif output.status_code == 200:

                if self.unpack(request, output) and 'view' in kw:
                    # Handling of @unpack and UnpackTemplateResponse
                    kw['view'].template_name = output.template_name
                    kw['view'].request._feincms_extra_context.update(
                        output.context_data)

                else:
                    # If the response supports deferred rendering, render the
                    # response right now. We do not handle template response
                    # middleware.
                    if hasattr(output, 'render') and callable(output.render):
                        output.render()

                    self.rendered_result = mark_safe(
                        output.content.decode('utf-8'))

                self.rendered_headers = {}

                # Copy relevant headers for later perusal
                for h in ('Cache-Control', 'Last-Modified', 'Expires'):
                    if h in output:
                        self.rendered_headers.setdefault(
                            h, []).append(output[h])

        elif isinstance(output, tuple) and 'view' in kw:
            kw['view'].template_name = output[0]
            kw['view'].request._feincms_extra_context.update(output[1])

        else:
            self.rendered_result = mark_safe(output)

        return True  # successful

    def send_directly(self, request, response):
        mimetype = response.get('Content-Type', 'text/plain')
        if ';' in mimetype:
            mimetype = mimetype.split(';')[0]
        mimetype = mimetype.strip()

        return (
            response.status_code != 200 or
            request.is_ajax() or
            getattr(response, 'standalone', False) or
            mimetype not in ('text/html', 'text/plain'))

    def unpack(self, request, response):
        return getattr(response, '_feincms_unpack', False)

    def render(self, **kwargs):
        return getattr(self, 'rendered_result', '')

    def finalize(self, request, response):
        headers = getattr(self, 'rendered_headers', None)
        if headers:
            self._update_response_headers(request, response, headers)

    def _update_response_headers(self, request, response, headers):
        """
        Combine all headers that were set by the different content types
        We are interested in Cache-Control, Last-Modified, Expires
        """

        # Ideally, for the Cache-Control header, we'd want to do some
        # intelligent combining, but that's hard. Let's just collect and unique
        # them and let the client worry about that.
        cc_headers = set(('must-revalidate',))
        for x in (cc.split(",") for cc in headers.get('Cache-Control', ())):
            cc_headers |= set((s.strip() for s in x))

        if len(cc_headers):
            response['Cache-Control'] = ", ".join(cc_headers)
        else:   # Default value
            response['Cache-Control'] = 'no-cache, must-revalidate'

        # Check all Last-Modified headers, choose the latest one
        lm_list = [parsedate(x) for x in headers.get('Last-Modified', ())]
        if len(lm_list) > 0:
            response['Last-Modified'] = http_date(mktime(max(lm_list)))

        # Check all Expires headers, choose the earliest one
        lm_list = [parsedate(x) for x in headers.get('Expires', ())]
        if len(lm_list) > 0:
            response['Expires'] = http_date(mktime(min(lm_list)))

    @classmethod
    def app_reverse_cache_key(self, urlconf_path, **kwargs):
        return 'FEINCMS:%s:APPCONTENT:%s:%s' % (
            getattr(settings, 'SITE_ID', 0),
            get_language(),
            urlconf_path,
        )

    @classmethod
    def closest_match(cls, urlconf_path):
        try:
            page_class = cls.parent.field.remote_field.model
        except AttributeError:
            page_class = cls.parent.field.rel.to

        contents = cls.objects.filter(
            parent__in=page_class.objects.active(),
            urlconf_path=urlconf_path,
        ).order_by('pk').select_related('parent')

        if len(contents) > 1:
            try:
                current = short_language_code(get_language())
                return [
                    content for content in contents if
                    short_language_code(content.parent.language) == current
                ][0]

            except (AttributeError, IndexError):
                pass

        try:
            return contents[0]
        except IndexError:
            pass

        return None



"""
Add a "featured" field to objects so admins can better direct top content.
"""

from __future__ import absolute_import, unicode_literals

from django.db import models
from django.utils.translation import ugettext_lazy as _

from feincms import extensions


class Extension(extensions.Extension):
    def handle_model(self):
        self.model.add_to_class(
            'featured',
            models.BooleanField(
                _('featured'),
                default=False,
            ),
        )

    def handle_modeladmin(self, modeladmin):
        modeladmin.add_extension_options(_('Featured'), {
            'fields': ('featured',),
            'classes': ('collapse',),
        })



# flake8: noqa
from __future__ import absolute_import, unicode_literals

import warnings

from feincms.apps import *

warnings.warn(
    'Import ApplicationContent and friends from feincms.apps.',
    DeprecationWarning, stacklevel=2)
